-------------------------------------------------------------------------------
-- Lorewalkers by Lockslap, US-Bleeding Hollow
-- <Borderline Amazing>, http://ba-guild.com
-------------------------------------------------------------------------------
Lorewalkers	= LibStub("AceAddon-3.0"):NewAddon("Lorewalkers", "AceEvent-3.0", "AceConsole-3.0")
if not TomTom then return end

-- libaries
local L					= LibStub("AceLocale-3.0"):GetLocale("Lorewalkers", false)
local Astrolabe			= DongleStub("Astrolabe-1.0")
local AceConfig			= LibStub("AceConfig-3.0")
local AceConfigDialog	= LibStub("AceConfigDialog-3.0")
local AceConfigRegistry	= LibStub("AceConfigRegistry-3.0")
local LSM				= LibStub("LibSharedMedia-3.0")

-- setup some globals
Lorewalkers.name 	= GetAddOnMetadata("Lorewalkers", "Title")
Lorewalkers.version	= GetAddOnMetadata("Lorewalkers", "Version")

local waypoint
local zones, db, waypoints	= {}, {}, {}

local enabledString		= ("|cff00ff00%s|r"):format(L["Enabled"])
local disabledString	= ("|cffff0000%s|r"):format(L["Disabled"])
local repStandings 		= {
	[1]	= L["Hated"],
	[2]	= L["Hostile"],
	[3]	= L["Unfriendly"],
	[4]	= L["Neutral"],
	[5]	= L["Friendly"],
	[6]	= L["Honored"],
	[7]	= L["Revered"],
	[8]	= L["Exalted"],
}

-- options page
local fields = {"Title", "Version", "Author", "X-Category", "X-License", "X-Email", "Email", "eMail", "X-Website", "X-Credits", "X-Localizations", "X-Donate", "X-Paypal"}
local fieldName = "|cff1784d1%s:|r"
local defaults 		= {
	char = {
		enabled 		= true,
		ssFrame			= true,
		debug			= UnitName("player") == "Lockslap" and true or false,	-- i need debugging by default
		notices			= UnitName("player") == "Lockslap" and true or false,	-- ditto for notices
		lockslap		= false,
		arrow			= true,
		minimap			= true,
		worldmap		= true,
		firstLoad		= false,	-- for building our waypoint db
		waypoints		= {},
	}
}
local options = {
	type	= "group",
	args	= {
		general	= {
			type	= "group",
			name	= Lorewalkers.name,
			get		= function(info) return db[info[#info]] end,
			set		= function(info, value) 
				db[info[#info]] = value
				if waypoint then
					TomTom:RemoveWaypoint(waypoint)
				end
				Lorewalkers:AddClosestWaypoint()
			end,
			args	= {
				header = {
					type		= "header",
					order		= 1,
					name		= L["Version"] .. (": |cff99ff33%s|r"):format(Lorewalkers.version),
					width		= "full",
				},
				description = {
					type		= "description",
					order		= 2,
					name		= L["A simple addon to aid you with obtaining exalted with The Lorewalkers."],
					width		= "full",
				},
				enabled = {
					type		= "toggle",
					order		= 3,
					name		= L["Enable Lorewalkers"],
					desc		= L["Toggle the waypoints."],
				},
				ssFrame = {
					type		= "toggle",
					order		= 4,
					name		= L["Screenshot Frame"],
					desc		= L["Toggle the use of the screenshot frame to show you the location of the criteria."],
				},
				debug = {
					type		= "toggle",
					order		= 5,
					name		= L["Debugging"],
					desc		= L["This will ignore if you have completed the achievements/criteria and load the waypoints.  I use this for fixing issues."],
				},
				notices = {
					type		= "toggle",
					order		= 6,
					name		= L["Display Notices"],
					desc		= L["Notify you of various actions being taken by the addon.  For example, loading a new waypoint."],
				},
				dummy = {
					type		= "description",
					order		= 99,
					name		= "",
					width		= "full",
				},
				addTo = {
					type		= "group",
					order		= 100,
					guiInline	= true,
					name		= L["Add To"],
					disabled	= function() return not db.enabled end,
					args		= {
						arrow = {
							type		= "toggle",
							order		= 1,
							name		= L["Crazy Arrow"],
							desc		= L["Add waypoints to TomTom's crazy arrow."],
							get			= function() return db.arrow end,
							set			= function(_, value)
								db.arrow = value
								if waypoint then
									TomTom:RemoveWaypoint(waypoint)
								end
								Lorewalkers:AddClosestWaypoint()
							end,
						},
						minimap = {
							type		= "toggle",
							order		= 2,
							name		= L["Minimap"],
							desc		= L["Add waypoints to the minimap."],
							get			= function() return db.minimap end,
							set			= function(_, value)
								db.minimap = value
								if waypoint then
									TomTom:RemoveWaypoint(waypoint)
								end
								Lorewalkers:AddClosestWaypoint()
							end,
						},
						worldmap = {
							type		= "toggle",
							order		= 3,
							name		= L["World Map"],
							desc		= L["Add waypoints to the world map."],
							get			= function() return db.worldmap end,
							set			= function(_, value)
								db.worldmap = value
								if waypoint then
									TomTom:RemoveWaypoint(waypoint)
								end
								Lorewalkers:AddClosestWaypoint()
							end,
						},
					},
				},
				ssFrameOptions = {
					type		= "group",
					order		= 101,
					guiInline	= true,
					name		= L["Screenshot Frame"],
					disabled	= function() return db.ssFrame end,
					args		= {
					
					},
				},
			},
		},
	}
}

local function InArray(needle, haystack)
	if #haystack == 0 then return false end
	for i = 1, #haystack do
		if haystack[i] == needle then
			return true
		end
	end
	return false
end

local function RepStatus()
	-- id 1345 are the lorewalkers
	local _, _, standing, barMin, barMax, barValue, _, _, _, _, _, _, _ = GetFactionInfoByID(1345)
	barValue = barValue - barMin
	barMax = barMax - barMin
	return barValue, barMax, repStandings[standing], math.floor(barValue / barMax * 100)
end

local function GetCriteriaCoordinates(criteria)
	for zone, waypoints in pairs(Lorewalkers.waypoints) do
		for name, info in pairs(waypoints) do
			if criteria == name then
				return zone, info.x, info.y
			end
		end
	end
	return false
end

function Lorewalkers:AddAllWaypoints()
	if GetCurrentMapContinent() ~= 6 or not db.waypoints[GetCurrentMapAreaID()] then return end
	local map, opts = GetCurrentMapAreaID(), {
		title			= "",
		minimap			= false,
		crazy			= true,
		world			= false,
		persistent		= false,
		cleardistance	= 0,
		arrivaldistance	= 0,
		callbacks		= {},
	}
	for _, point in pairs(db.waypoints[map]) do
		opts.title = ("%s\n%s"):format(point.name, point.achieveName)
		waypoints[#waypoints + 1] = TomTom:AddMFWaypoint(map, 0, point.x / 100, point.y / 100, opts)
	end
end

function Lorewalkers:ClearAllWaypoints()
	if not waypoints or #waypoints == 0 then return end
	for i = 1, #waypoints do
		TomTom:RemoveWaypoint(waypoints[i])
	end
end

function Lorewalkers:UpdateCriteria()
	for mapID, waypoints in pairs(db.waypoints) do
		for cID, point in pairs(waypoints) do
			local _, _, completed, _, _, _, _, _, _, _ = GetAchievementCriteriaInfo(point.achieveID, point.num)
			db.waypoints[mapID][cID].completed = completed
		end
	end
end

function Lorewalkers:HandleChatCommand(args)
	local command, value = self:GetArgs(args, 2)
	local iconString = "\124T%s:12\124t"
	local completedString = "%s%s %s |cff1784d1%d/%d/%d|r"
	local notCompletedString = "%s%s |cffff0000%s|r"
	local criteriaCompleted = "    %d. %s: |cffffff00%s|r%s |cff00ff00%s|r"
	local criteriaNotCompleted = "    %d. %s: |cffffff00%s|r%s |cffff0000%s|r"
	local achievementPattern = "\124c%x+\124Hachievement:(%d-):.-\124h%[.-%]\124h\124r"
	
	if command == "toggle" then
		db.enabled = not db.enabled
		Lorewalkers:Print((L["Waypoints have been %s"]):format(db.enabled == true and enabledString or disabledString))
		if not db.enabled and waypoint then
			TomTom:RemoveWaypoint(waypoint)
		else
			Lorewalkers:UpdateCriteria()
			Lorewalkers:AddClosestWaypoint()
		end
		
	elseif command == "config" then
		InterfaceOptionsFrame_OpenToCategory(self.configFrame)
		
	elseif command == "rep" or command == "reputation" then
		local exaltedString	= "%s |cff1784d1%d/%d %s (%d%%)|r %s |cff1784d1%s|r."
		local value, max, standing, percent = RepStatus()
		self:Print(exaltedString:format(L["You are"], value, max, standing, percent, L["with"], L["The Lorewalkers"]))
		
	elseif command == "criterias" or command == "criteria" or value:match(achievementPattern) then
		local id = 0
		if value:match(achievementPattern) then
			id = value:match(achievementPattern)
		elseif value:find(" ") then
			-- they provided a string, it must be in quotes
			if not Lorewalkers.achievements[value] then
				self:Print((L["Achievement |cffff0000\"%s\"|r doesn't exist."]):format(value))
				return
			else
				id = Lorewalkers.achievements[value]
			end
		else
			id = tonumber(value)
		end
		local link = GetAchievementLink(id)
		local _, _, _, completed, month, day, year, _, _, icon, _, _ = GetAchievementInfo(id)
		self:Print(completed == true and completedString:format(iconString:format(icon), link, L["Completed on"], month, day, year) or notCompletedString:format(iconString:format(icon), link, L["Incomplete"]))
		for i = 1, GetAchievementNumCriteria(id) do
			local description, _, completed, _, _, _, _, _, _, _ = GetAchievementCriteriaInfo(id, i)
			local zone, x, y = GetCriteriaCoordinates(description)
			if zone then
				print(completed == true and criteriaCompleted:format(i, description, zone, ("(%.2f, %.2f)"):format(x, y), L["Completed"]) or criteriaNotCompleted:format(i, description, zone, ("(%.2f, %.2f)"):format(x, y), L["Incomplete"]))
			else
				print(completed == true and criteriaCompleted:format(i, description, "", "", L["Completed"]) or criteriaNotCompleted:format(i, description, "", "", L["Incomplete"]))
			end
		end
		
	elseif command == "achieve" or command == "achievements" then
		for _, id in pairs(self.achievements) do
			local link = GetAchievementLink(id)
			local _, _, _, completed, month, day, year, _, _, icon, _, _ = GetAchievementInfo(id)
			self:Print(completed == true and completedString:format(iconString:format(icon), link, L["Completed on"], month, day, year) or notCompletedString:format(iconString:format(icon), link))
		end
		
	elseif command == "clear" then
		if waypoint then TomTom:RemoveWaypoint(waypoint) end
		self:ClearAllWaypoints()	-- for "lockslap" mode
	
	elseif command == "debug" then
		db.debug = not db.debug
		self:Print((L["Debugging has been %s"]):format(db.debug == true and enabledString or disabledString))
		self:UpdateWaypoints()
	
	-- my own command to make taking screenshots easier
	-- this simply ignores if waypoints are completed and loads them all
	elseif command == "lockslap" then
		db.lockslap = not db.lockslap
		self:Print((L["Lockslap mode %s"]):format(db.lockslap == true and enabledString or disabledString))
		if db.lockslap then
			db.enabled = false
			db.debug = false
			if GetCurrentMapContinent() == 6 then
				self:AddAllWaypoints()
			end
		else
			self:ClearAllWaypoints()
		end
		
	elseif not command or command == "" or command == "help" or command == "?" then
		self:Print((L["Lorewalkers version |cff1784d1%.2f|r by |cff00ff00Lockslap|r"]):format(self.version))
		self:Print(L["Chat Command:  |cff1784d1/lorewalkers|r or |cff1784d1/lw|r"])
		print(L["    |cff1784d1config|r - Open Configuration Window"])
		print(L["    |cff1784d1clear|r - Clear Waypoints"])
		print(L["    |cff1784d1help|r or |cff1784d1?|r - Print This Help Again"])
		print(L["    |cff1784d1toggle|r - Toggle the Waypoints"])
		print(L["    |cff1784d1rep|r or |cff1784d1reputation|r - Your Reputation with the Lorewalkers"])
		print(L["    |cff1784d1criteria|r or |cff1784d1criterias|r |cff1784d1<id or name>|r - Display Achievement Criteria"])
		print(L["    |cff1784d1achievements|r or |cff1784d1achieve|r - Lorewalkers Achievement Progress"])
		
	end
end

function Lorewalkers:GetClosestWaypoint()
	local distance, point, criteriaID
	local mapID = GetCurrentMapAreaID()
	local map, level, x, y = Astrolabe:GetCurrentPlayerPosition()
	if not db.waypoints[mapID] then return false end
	for criteria, points in pairs(db.waypoints[mapID]) do
		local wayDistance = Astrolabe:ComputeDistance(map, level, x, y, mapID, 0, points.x / 100, points.y / 100)
		if (db.debug and wayDistance > 50) or not points.completed then
			if not distance or wayDistance < distance then
				distance = wayDistance
				point = points
				criteriaID = criteria
			end
		end
	end
	-- we should have our nearest waypoint
	if not distance or not point then
		return false
	else
		return mapID, criteriaID, point, distance
	end
end

function Lorewalkers:AddClosestWaypoint()
	local zone = GetZoneText()
	if db.debug and GetCurrentMapContinent() == 6 and not db.waypoints[GetCurrentMapAreaID()] then
		self:Print((L["No waypoints found for |cff1784d1%s|r."]):format(zone))
	elseif GetCurrentMapContinent() == 6 and (db.debug or db.enabled) then
		local mapID, criteriaID, waypoint, distance = self:GetClosestWaypoint()
		if not mapID or not criteriaID or not waypoint or not distance then
			if db.notices then
				self:Print((L["You have completed all waypoints for |cff1784d1%s|r"]):format(zone))
			end
		else
			if db.debug or db.notices then
				self:Print((L["Loaded waypoint for |cff1784d1%s|r, |cff1784d1%.2f|r yards away."]):format(waypoint.name, distance))
			end
		
			-- thanks to Cladhaire, the author of TomTom, for pointing me in the right direction
			local opts = {
				title			= ("%s\n%s"):format(waypoint.name, waypoint.achieveName),
				persistent		= false,
				minimap			= db.minimap,
				world			= db.worldmap,
				crazy			= db.arrow,
				cleardistance	= 0,
				arrivaldistance	= 0,
				callbacks		= {},
			}
			opts.callbacks = TomTom:DefaultCallbacks(opts)
			opts.callbacks.distance[15] = function(event, uid, range, distance, lastDistance)
				if lastDistance > distance then
					Lorewalkers:UpdateCriteria()
					TomTom:RemoveWaypoint(uid)
					Lorewalkers:AddClosestWaypoint()
				end
			end
			waypoint = TomTom:AddMFWaypoint(mapID, 0, waypoint.x / 100, waypoint.y / 100, opts)
		end
	end
end

function Lorewalkers:HandleEvents(event, ...)
	if waypoint then
		TomTom:RemoveWaypoint(waypoint)
		waypoint = nil
	end
	
	if not db.lockslap then
		if GetCurrentMapContinent() == 6 and (db.debug or db.enabled) then
			self:AddClosestWaypoint()
		end
	else
		self:ClearAllWaypoints()
		self:AddAllWaypoints()
	end
end
Lorewalkers:RegisterEvent("ZONE_CHANGED_NEW_AREA", "HandleEvents")

function Lorewalkers:OnInitialize()
	-- load the config
	self.db = LibStub("AceDB-3.0"):New("LorewalkersDB", defaults)
	db = self.db.char
	
	-- setup tomtom
	TomTom.profile.arrow.enable = true
	TomTom.profile.arrow.arrival = 5
	
	-- this will be done the first time the addon is loaded
	if not db.firstLoad then
		for mapID, waypoints in pairs(self.waypoints) do
			for cID, point in pairs(waypoints) do
				local _, achieveName, _, _, _, _, _, _, _, _, _, _, _ = GetAchievementInfo(point.achieveID)
				for i = 1, GetAchievementNumCriteria(point.achieveID) do
					local description, _, completed, _, _, _, _, _, _, criteriaID = GetAchievementCriteriaInfo(point.achieveID, i)
					if cID == criteriaID then
						if not db.waypoints[mapID] then db.waypoints[mapID] = {} end
						if not db.waypoints[mapID][cID] then db.waypoints[mapID][cID] = {} end
						db.waypoints[mapID][cID] = {
							achieveID	= point.achieveID,
							achieveName	= achieveName,
							name		= description,
							num			= i,
							x			= point.x,
							y			= point.y,
							completed	= completed,
						}
					end
				end
			end
		end
		db.firstLoad = true
	else
		self:UpdateCriteria()
	end
end

function Lorewalkers:OnEnable()
	-- add the about group
	options.args.about = {
		type		= "group",
		order		= 101,
		name		= L["About"],
		args		= {},
	}
	local i = 1
	for _, field in pairs(fields) do
		local val = GetAddOnMetadata(self.name, field)
		if val then
			options.args.about.args[field] = {
				type	= "description",
				order	= i,
				width	= "normal",
				name	= fieldName:format(field:gsub("X%-", "")),
			}
			i = i + 1
			options.args.about.args[("%s_value"):format(field)] = {
				type	= "description",
				order	= i,
				width	= "double",
				name	= val,
			}
			i = i + 1
		end
	end

	-- setup config window
	AceConfigRegistry:RegisterOptionsTable(self.name, options)
	self.frames = {}
	self.frames.general = AceConfigDialog:AddToBlizOptions(self.name, nil, nil, "general")
	self.frames.about = AceConfigDialog:AddToBlizOptions(self.name, L["About"], self.name, "about")

	self:RegisterChatCommand("lorewalkers", "HandleChatCommand")
	self:RegisterChatCommand("lw", "HandleChatCommand")
	
	-- pandaria's continent id is 6
	if GetCurrentMapContinent() == 6 and (db.debug or db.enabled) then
		self:AddClosestWaypoint()
	end
end

function Lorewalkers:OnDisable()
	self:UnregisterChatCommand("lorewalkers")
	self:UnregisterChatCommand("lw")
	
	if waypoint then
		TomTom:RemoveWaypoint(waypoint)
		waypoint = nil
	end
end